# -*- coding: utf-8 -*-
"""
Created on Sun Jun 22 12:41:08 2025

@author: Thomas Lee
Rice University
Department of Earth, Environmental, and Planetary Sciences
Email: tl165@rice.edu

AviList Citation:
AviList Core Team. 2025. AviList: The Global Avian Checklist, v2025. https://doi.org/10.2173/avilist.v2025
"""

from AviListPy.data.avilistdatabase import AviListDataBase
from AviListPy.taxonomy.species import Species
from typing import Any, KeysView, ValuesView, ItemsView, Iterator, List
from pandas import DataFrame

class Genus():
    """Container for a Genus in the AviList DataBase

    The third lowest/highest taxonomic rank in the AviList.taxonomy class
    system. Container for all Species objects corresponding with species
    in this genus. See Genus.keys() for a list of all available data. Can
    access data in a dictionary-like manner where keys are columns in the
    AviList, but iterating will iterate through Genus.species

    Attributes:
    -----------
    db: AviList.data.avilistdatabase.AviListDataBase
        AviListDataBase class. It is recommended to pass an existing
        AviListDataBase object to the Genus class during initialization,
        but if none is given it will initialize one from the Excel sheet. See
        the Setup section on the main GitHub page for more detail.
    df: Pandas.DataFrame
        The single row for this genus in AviList as a Pandas DataFrame.
    name: str
        Scientific name for this genus, from self['Scientific_name']
    order: str
        Taxonomic order
    family: str
        Taxonomic family
    species: list of AviList.taxonomy.species.Species, or None

    Example:
        >>> db = AviListDataBase()
        >>> genus = Genus('Ardea',db=db)
        >>> genus.name
        'Ardea'
        >>> genus['Scientific_name']
        'Ardea'
    """
    def __init__(self, name: str, exact: bool=False, load_subspecies: bool=False, db: AviListDataBase=None):
        """
        Parameters
        ----------
        name : str
            Scientific name of the species to search for.
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any English name
            in the database, and is not case sensitive. The default is False.
        load_subspecies: bool, optional
            If True, will load Subspecies objects while loading Species. See
            AviList.taxonomy.species.Species for more information.
        db : AviListDataBase, optional
            AviListDataBase. The default is None.
        """
        if db is None:
            self.db = AviListDataBase()
        else:
            self.db = db
        self.df = self.lookup_genus(name, exact=exact)
        self._data = self.df.iloc[0].to_dict()
        self.name = self._data['Scientific_name']
        self.family = self._data['Family']
        self.order = self._data['Order']
        self.species = self.find_matching_species(load_subspecies=load_subspecies)

    def __str__(self) -> str:
        return_str = f'{self["Scientific_name"]}'
        num_equals = (80 - len(return_str)) // 2
        return_str = '='*num_equals + return_str + '='*num_equals
        for key, val in self.items():
            return_str += (f'\n{key}: {val}')
        return return_str + '\n'

    def __iter__(self) -> Iterator[Any]:
        return iter(self.species)

    def __len__(self) -> int:
        return len(self.species)

    def __getitem__(self, key) -> Any:
        return self._data[key]

    def __setitem__(self, key, value) -> None:
        self._data[key] = value

    def __contains__(self, key) -> bool:
        return key in self._data

    def keys(self) -> KeysView:
        """Returns keys in a dictionary.keys() like manner"""
        return self._data.keys()

    def values(self) -> ValuesView:
        """Returns values in a dictionary.values() like manner"""
        return self._data.values()

    def items(self) -> ItemsView:
        """Returns keys, values in a dictionary.items() like manner"""
        return self._data.items()

    def lookup_genus(self, name, exact: bool=False) -> DataFrame:
        """
        Parameters
        ----------
        name : str
            Genus to search for.
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any scientific name
            in the database, and is not case sensitive. The default is False.

        Returns
        -------
        _genus_df : pandas.DataFrame
           Pandas DataFrame with only one row containing the entry for the genus.

        """
        df = self.db.df[self.db.df['Taxon_rank'].str.contains('genus')]
        if exact is True:
            _genus_df = df[df['Scientific_name'] == name]
        else:
            _genus_df = df[df['Scientific_name'].str.contains(name, case=False, na=False)]

        if _genus_df.shape[0] == 0:
            raise ValueError('No matching species found')
        if _genus_df.shape[0] > 1:
            fail_str = f'{name} could refer to: \n'
            for _genus_ in _genus_df['Scientific_name'].to_list():
                fail_str += (f'{_genus_}, ')
            raise ValueError(fail_str)
        _genus_df = _genus_df.dropna(axis=1)
        return _genus_df

    def find_matching_species(self, load_subspecies: bool=False) -> List[Species]:
        """
        Parameters
        ----------
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any scientific name
            in the database, and is not case sensitive. The default is False.
        load_subspecies : bool, optional
            If True, loads subspecies. The default is False.

        Returns
        -------
        matching_species_list : list of AviList.taxonomy.species.Species
            List of species contained within the order.
        """
        species_df = self.db.df[self.db.df['Taxon_rank'] == 'species']
        matching_species_df = species_df[species_df['Scientific_name'].str.contains(self.name)]
        matching_species_list = []
        for _matching_species_name in matching_species_df['English_name_AviList'].to_list():
            matching_species_list.append(Species(_matching_species_name, db = self.db, exact=True, load_subspecies=load_subspecies))
        return matching_species_list

    def show_species(self) -> None:
        """Prints each species in this genus"""
        for species in self.species:
            print(species)



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 22 13:14:46 2025

@author: Thomas Lee
Rice University
Department of Earth, Environmental, and Planetary Sciences
Email: tl165@rice.edu

AviList Citation:
AviList Core Team. 2025. AviList: The Global Avian Checklist, v2025. https://doi.org/10.2173/avilist.v2025
"""

from AviListPy.data.avilistdatabase import AviListDataBase
from AviListPy.taxonomy.genus import Genus, Species
from typing import Any, KeysView, ValuesView, ItemsView, Iterator, List
from pandas import DataFrame

class Family():
    """Container for a Family in the AviList DataBase

    The second highest taxonomic rank in the AviList.taxonomy class
    system. Container for all Genus objects corresponding with genera
    in this family. See Family.keys() for a list of all available data. Can
    access data in a dictionary-like manner where keys are columns in the
    AviList, but iterating will iterate through Family.genera

    Attributes:
    -----------
    db: AviList.data.avilistdatabase.AviListDataBase
        AviListDataBase class. It is recommended to pass an existing
        AviListDataBase object to the Family class during initialization,
        but if none is given it will initialize one from the Excel sheet. See
        the Setup section on the main GitHub page for more detail.
    df: Pandas.DataFrame
        The single row for this family in AviList as a Pandas DataFrame.
    name: str
        Scientific name for this family, from self['Scientific_name']
    order: str
        Taxonomic order
    family: str
        Taxonomic family
    genera: list of AviList.taxonomy.genus.Genus
        All genera within this family.
    species: list of AviList.taxonomy.species.Species
        All species within this family.

    Example:
        >>> db = AviListDataBase()
        >>> family = Family('Anatidae',db=db)
        >>> family.name
        'Anatidae'
        >>> family['Family_English_name']
        'Ducks, Swans, and Geese'

    """
    def __init__(self, name: str, exact: bool=False, load_subspecies: bool=False, db: AviListDataBase=None):
        """
        Parameters
        ----------
        name : str
            Scientific name of the family to search for.
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any English name
            in the database, and is not case sensitive. The default is False.
        load_subspecies: bool, optional
            If True, will load Subspecies objects while loading Species. See
            AviList.taxonomy.species.Species for more information.
        db : AviListDataBase, optional
            AviListDataBase. The default is None.
        """
        if db is None:
            self.db = AviListDataBase()
        else:
            self.db = db
        self.df = self.lookup_family(name, exact=exact)
        self._data = self.df.iloc[0].to_dict()
        self.name = self._data['Scientific_name']
        self.order = self._data['Order']
        self.genera = self.find_matching_genera(load_subspecies=load_subspecies)
        self.species = self.find_matching_species()

    def __str__(self) -> str:
        return_str = f'{self["Scientific_name"]}'
        num_equals = (80 - len(return_str)) // 2
        return_str = '='*num_equals + return_str + '='*num_equals
        for key, val in self.items():
            return_str += (f'\n{key}: {val}')
        return return_str + '\n'

    def __iter__(self) -> Iterator[Any]:
        return iter(self.genera)

    def __len__(self) -> int:
        return len(self.genera)

    def __getitem__(self, key) -> Any:
        return self._data[key]

    def __setitem__(self, key, value) -> None:
        self._data[key] = value

    def __contains__(self, key) -> bool:
        return key in self._data

    def keys(self) -> KeysView:
        """Returns keys in a dictionary.keys() like manner"""
        return self._data.keys()

    def values(self) -> ValuesView:
        """Returns values in a dictionary.values() like manner"""
        return self._data.values()

    def items(self) -> ItemsView:
        """Returns keys, values in a dictionary.items() like manner"""
        return self._data.items()

    def lookup_family(self, name, exact: bool=False) -> DataFrame:
        """
        Parameters
        ----------
        name : str
            Family to search for.
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any scientific name
            in the database, and is not case sensitive. The default is False.

        Returns
        -------
        _family_df : pandas.DataFrame
           Pandas DataFrame with only one row containing the entry for the family.

        """
        df = self.db.df[self.db.df['Taxon_rank'].str.contains('family')]
        if exact is True:
            _family_df = df[df['Scientific_name'] == name]
        else:
            _family_df = df[df['Scientific_name'].str.contains(name, case=False, na=False)]
        if _family_df.shape[0] == 0:
            raise ValueError('No matching families found')
        if _family_df.shape[0] > 1:
            fail_str = f'{name} could refer to: \n'
            for _family_ in _family_df['Scientific_name'].to_list():
                fail_str += (f'{_family_}, ')
            raise ValueError(fail_str)
        _family_df = _family_df.dropna(axis=1)
        return _family_df

    def find_matching_genera(self, load_subspecies: bool=False) -> List[Genus]:
        """
        Parameters
        ----------
            If True, loads subspecies. The default is False.

        Returns
        -------
        matching_genera_list : list of AviList.taxonomy.genus.Genus
            List of genera contained within the order.
        """
        genus_df = self.db.df[self.db.df['Taxon_rank'] == 'genus']
        matching_genus_df = genus_df[genus_df['Family'] == str(self.name)]
        print(f'Loading {len(matching_genus_df)} genera in family {self.name}')
        if len(matching_genus_df) == 0:
            raise ValueError('No matching genera found')
        matching_genera_list = []
        for _matching_genus_name in matching_genus_df['Scientific_name'].to_list():
            matching_genera_list.append(Genus(_matching_genus_name, db=self.db, exact=True, load_subspecies=load_subspecies))
        return matching_genera_list

    def find_matching_species(self) -> List[Species]:
        """Returns a list of all species in this family as Species objects"""
        species_list = []
        for genus in self.genera:
            species_list += genus.species
        return species_list

    def show_genera(self) -> None:
        print(f'{len(self.genera)} genera in family {self.name}')
        count = 0
        for genus in self.genera:
            print(f'{genus.name}: {len(genus.species)} species')
            count += len(genus.species)
        print(f'{count} total species in {self.name}')



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 22 12:54:54 2025

@author: Thomas Lee
Rice University
Department of Earth, Environmental, and Planetary Sciences
Email: tl165@rice.edu
"""

all = ['order','family','genus','species','subspecies']


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 22 11:31:20 2025

@author: Thomas Lee
Rice University
Department of Earth, Environmental, and Planetary Sciences
Email: tl165@rice.edu

AviList Citation:
AviList Core Team. 2025. AviList: The Global Avian Checklist, v2025. https://doi.org/10.2173/avilist.v2025
"""
import os
from importlib import resources
from datetime import datetime
import pickle
import pandas as pd

from pandas import DataFrame

class AviListDataBase():
    def __init__(self, path: str=None, overwrite_existing=False, verbose=False) -> None:
        def verbose_print(*args, **kwargs):
            """Helper function that only prints if verbose is True"""
            if verbose:
                print(*args, **kwargs)

        if overwrite_existing is True or path is None:
            if path is not None and os.path.exists(path) is False:
                print(f'Could not find database at {path}. Creating a new database, and saving it there')
            verbose_print('Loading database...')
            start_time = datetime.now()
            self.df = self.load_df()
            verbose_print(f'Database loaded in {datetime.now() - start_time}')
            if path is not None:
                self._save(path)
        else:
            start_time = datetime.now()
            try:
                with open(path, 'rb') as f:
                    saved_version = pickle.load(f)
                self.__dict__.update(saved_version.__dict__)
                verbose_print(f'Database loaded in {datetime.now() - start_time}')
            except FileNotFoundError:
                verbose_print('Could not find pickled database, initializing new...')
                self.df = self.load_df()
                verbose_print(f'Database loaded in {datetime.now() - start_time}')

    def load_df(self) -> pd.DataFrame:
        """Loads the AviList Excel file as a Pandas DataFrame"""
        with resources.files('AviListPy.data').joinpath('AviList-v2025-11Jun-extended.xlsx').open('rb') as f:
            return pd.read_excel(f)

    def _save(self, path=None) -> None:
        """Pickles the AviListDataBase to the disk"""
        if path is None:
            raise ValueError('Must define path to save AviListDataBase.db to')
        try:
            with open(path, 'wb') as file:
                pickle.dump(self, file)
        except FileNotFoundError as e:
            raise ValueError(f'FileNotFoundError raised. Likely that the parent directory for your database save location does not exist: \n: {str(e)}')

    def keys(self):
        """Returns all columns in the data base"""
        return self.df.columns.tolist()

    def query(self, search, category: str=None, exact: bool=False) -> DataFrame:
        if exact is True:
            return_df = self._exact_search(search=search,category=category)
        else:
            return_df = self._substring_search(search=search,category=category)

        print(f'Search resulted in {len(return_df)} entries')

        return return_df

    def _exact_search(self, search: str, category: str) -> DataFrame:
        try:
            if category is None:
                return_df = self.df[self.df.isin([search]).any(axis=1)]
            else:
                return_df = self.df[self.df[category] == search]
        except KeyError:
            raise KeyError(f'Invalid category. Valid categories are {self.keys()}')

        return return_df

    def _substring_search(self, search: str, category: str) -> DataFrame:
        try:
            if category is None:
                df_str = self.df.astype(str)
                mask = df_str.apply(lambda col: col.str.contains(search, na=False, case=False)).any(axis=1)
                return_df = self.df[mask]
            else:
                return_df = self.df[self.df[category].str.contains(search, case=False, na=False)]
        except KeyError:
            raise KeyError(f'Invalid category. Valid categories are {self.keys()}')

        return return_df


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 22 13:58:39 2025

@author: Thomas Lee
Rice University
Department of Earth, Environmental, and Planetary Sciences
Email: tl165@rice.edu

AviList Citation:
AviList Core Team. 2025. AviList: The Global Avian Checklist, v2025. https://doi.org/10.2173/avilist.v2025
"""

from AviListPy.data.avilistdatabase import AviListDataBase
from typing import Any, KeysView, ValuesView, ItemsView
from pandas import DataFrame

class Subspecies():
    """Container for a Subspecies in the AviList DataBase

    The lowest taxonomic rank in the AviList.taxonomy class system. Generally,
    this class is better accessed by initializing the Species class and
    using the load_subspecies=True option, and then accessing this class inside
    the Species.subspecies attribute

    Attributes:
    -----------
    db: AviList.data.avilistdatabase.AviListDataBase
        AviListDataBase class. It is recommended to pass an existing
        AviListDataBase object to the Subspecies class during initialization,
        but if none is given it will initialize one from the Excel sheet. See
        the Setup section on the main GitHub page for more detail.
    df: Pandas.DataFrame
        The single row for this subspecies in AviList as a Pandas DataFrame.
    name: str
        Scientific name for this subspecies, from self['Scientific_name']
    order: str
        Taxonomic order
    family: str
        Taxonomic family
    genus: str
        Taxonomic genus

    Example:
        >>> db = AviListDataBase()
        >>> subspecies = Subspecies(name = "Ardea alba egretta",db=db)
        >>> subspecies.name
        'Adra alba egretta'
    """
    def __init__(self, name, exact: bool=False, db: AviListDataBase=None):
        """
        Parameters
        ----------
        name : str
            Scientific name of the subspecies to search for.
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any scientific name
            in the database, and is not case sensitive. The default is False.
        db : AviListDataBase, optional
            AviListDataBase. The default is None.
        """
        if db is None:
            self.db = AviListDataBase()
        else:
            self.db = db
        self.df = self.lookup_subspecies(name, exact=exact)
        self._data = self.df.iloc[0].to_dict()
        self.name = self._data['Scientific_name']
        self.order = self._data['Order']
        self.family = self._data['Family']
        self.genus = self.get_genus()

    def __str__(self) -> str:
        return_str = f'{self["Scientific_name"]}'
        num_equals = (80 - len(return_str)) // 2
        return_str = '='*num_equals + return_str + '='*num_equals
        for key, val in self.items():
            return_str += (f'\n{key}: {val}')
        return return_str + '\n'

    def __getitem__(self, key) -> Any:
        return self._data[key]

    def __setitem__(self, key, value) -> None:
        self._data[key] = value

    def __contains__(self, key) -> bool:
        return key in self._data

    def keys(self) -> KeysView:
        """Returns keys in a dictionary.keys() like manner"""
        return self._data.keys()

    def values(self) -> ValuesView:
        """Returns values in a dictionary.values() like manner"""
        return self._data.values()

    def items(self) -> ItemsView:
        """Returns keys, values in a dictionary.items() like manner"""
        return self._data.items()

    def lookup_subspecies(self, name: str, exact: bool=False) -> DataFrame :
        """
        Parameters
        ----------
        name : str
            Subspecies to search for.
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any scientific name
            in the database, and is not case sensitive. The default is False.

        Returns
        -------
        _subspecies_df : pandas.DataFrame
           Pandas DataFrame with only one row containing the entry for the subspecies.

        """
        df = self.db.df
        if exact is True:
            _subspecies_df = df[df['Scientific_name'] == name]
        else:
            _subspecies_df = df[df['Scientific_name'].str.contains(name, case=False, na=False)]

        if _subspecies_df.shape[0] == 0:
            raise ValueError('No matching species found')
        if _subspecies_df.shape[0] > 1:
            fail_str = f'{name} could refer to: \n'
            for _subspecies_ in _subspecies_df['English_name_AviList'].to_list():
                fail_str += (f'{_subspecies_}, ')
            raise ValueError(fail_str)
        _subspecies_df = _subspecies_df.dropna(axis=1)
        return _subspecies_df

    def get_genus(self) -> str:
        """Returns the genus as a string"""
        return self.name.split(' ')[0]



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 22 11:38:06 2025

@author: Thomas Lee
Rice University
Department of Earth, Environmental, and Planetary Sciences
Email: tl165@rice.edu

AviList Citation:
AviList Core Team. 2025. AviList: The Global Avian Checklist, v2025. https://doi.org/10.2173/avilist.v2025
"""

from AviListPy.data.avilistdatabase import AviListDataBase
from AviListPy.taxonomy.subspecies import Subspecies
from typing import Any, KeysView, ValuesView, ItemsView, List
from pandas import DataFrame

class Species():
    """Container for a Species in the AviList DataBase

    The second lowest taxonomic rank in the AviList.taxonomy class system. This
    class is initialized with the English name for the species. This will be
    updated to support initializing with the scientific name in the future.

    Attributes:
    -----------
    db: AviList.data.avilistdatabase.AviListDataBase
        AviListDataBase class. It is recommended to pass an existing
        AviListDataBase object to the Species class during initialization,
        but if none is given it will initialize one from the Excel sheet. See
        the Setup section on the main GitHub page for more detail.
    df: Pandas.DataFrame
        The single row for this subspecies in AviList as a Pandas DataFrame.
    name: str
        English name for this species, from self['English_name_AviList']
    scientific_name: str
        Scientific name for this subspecies, from self['Scientific_name']
    order: str
        Taxonomic order
    family: str
        Taxonomic family
    genus: str
        Taxonomic genus
    subspecies: list of AviList.taxonomy.subspecies.Subspecies, or None
        If load_subspecies is False, then None. If True, then list of
        Subspecies objects.

    Example:
        >>> db = AviListDataBase()
        >>> species = Species(name = "Great Egret",db=db)
        >>> subspecies.name
        'Great Egret'
        >>> subspecies['Bibliographic_details']
        'Systema Naturæ per Regna Tria Naturæ, Secundum Classes, Ordines,
        Genera, Species, cum Characteribus, Differentiis, Synonymis, Locis.
        Tomus I. Editio decima, reformata 1 p.144'
    """
    def __init__(self, name: str, exact: bool=False, load_subspecies=False, db: AviListDataBase=None):
        """
        Parameters
        ----------
        name : str
            English name of the species to search for.
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any English name
            in the database, and is not case sensitive. The default is False.
        load_subspecies: bool, optional
            If True, will load Subspecies objects for each subspecies and write
            them in a list to Species.subspecies
        db : AviListDataBase, optional
            AviListDataBase. The default is None.
        """
        if db is None:
            self.db = AviListDataBase()
        else:
            self.db = db
        self.df = self.lookup_species(name, exact=exact)
        self._data = self.df.iloc[0].to_dict()
        self.name = self._data['English_name_AviList']
        self.scientific_name = self._data['Scientific_name']
        self.order = self._data['Order']
        self.family = self._data['Family']
        self.genus = self.get_genus()
        self.subspecies = None

        if load_subspecies is True:
            self.subspecies = self.get_subspecies()

    def __str__(self) -> str:
        return_str = f'{self["English_name_AviList"]}'
        num_equals = (80 - len(return_str)) // 2
        return_str = '='*num_equals + return_str + '='*num_equals
        if self.subspecies is not None:
            return_str += f'\nContains {len(self.subspecies)} subspecies'
        for key, val in self.items():
            return_str += (f'\n{key}: {val}')
        return return_str + '\n'

    def __getitem__(self, key) -> Any:
        return self._data[key]

    def __setitem__(self, key, value) -> None:
        self._data[key] = value

    def __contains__(self, key) -> bool:
        return key in self._data

    def keys(self) -> KeysView:
        """Returns keys in a dictionary.keys() like manner"""
        return self._data.keys()

    def values(self) -> ValuesView:
        """Returns values in a dictionary.values() like manner"""
        return self._data.values()

    def items(self) -> ItemsView:
        """Returns keys, values in a dictionary.items() like manner"""
        return self._data.items()

    def lookup_species(self, name: str, exact: bool=False) -> DataFrame:
        """
        Parameters
        ----------
        name : str
            Species to search for.
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any scientific name
            in the database, and is not case sensitive. The default is False.

        Returns
        -------
        _species_df : pandas.DataFrame
           Pandas DataFrame with only one row containing the entry for the species.
        """
        df = self.db.df

        # First we search for a matching English name
        try:
            if exact is True:
                _species_df = df[df['English_name_AviList'] == name]
            else:
                _species_df = df[df['English_name_AviList'].str.contains(name, case=False, na=False)]

            if _species_df.shape[0] == 0:
                raise ValueError('No matching species found')
        except ValueError as e:
            # We only want to search for scientific name if the English name
            # search doesn't return any results
            if str(e) == 'No matching species found':
                # Scientific name will always be a substring of the subspecies
                # We need to limit our results to only species
                if exact is True:
                    _species_df = df[df['Scientific_name'] == name]
                    _species_df = _species_df[_species_df['Taxon_rank'] == 'species']
                else:
                    _species_df = df[df['Scientific_name'].str.contains(name, case=False, na=False)]
                    _species_df = _species_df[_species_df['Taxon_rank'] == 'species']

        if _species_df.shape[0] == 0:
            raise ValueError('No matching species found')

        if _species_df.shape[0] > 1:
            fail_str = f'{name} could refer to: \n'
            for _species_ in _species_df['English_name_AviList'].to_list():
                fail_str += (f'{_species_}, ')
            raise ValueError(fail_str)

        # Drop nan columns
        _species_df = _species_df.dropna(axis=1)

        return _species_df

    def get_subspecies(self) -> List[Subspecies]:
        """Pulls all subspecies for this species and writes them as a list of
        AviList.taxonomy.subspecies.Subspecies objects to Species.subspecies"""
        subspecies_df = self.db.df[self.db.df['Taxon_rank'] == 'subspecies']
        matching_subspecies_df = subspecies_df[subspecies_df['Scientific_name'].str.contains(self.scientific_name, case=False,na=False)]
        matching_subspecies_list = []
        for _matching_subspecies_name in matching_subspecies_df['Scientific_name'].to_list():
            matching_subspecies_list.append(Subspecies(_matching_subspecies_name, db=self.db, exact=True))
        return matching_subspecies_list

    def get_genus(self) -> str:
        """Returns the genus of this species as a string"""
        return self.scientific_name.split(' ')[0]

    def brief_summary(self) -> str:
        """Returns a short, easily readable version of info about this species"""
        return_str = f'{self["English_name_AviList"]}'
        num_equals = (80 - len(return_str)) // 2
        return_str = '='*num_equals + return_str + '='*num_equals
        return_str += f"\nScientific Name: {self['Scientific_name']}"
        if self.subspecies is not None:
            return_str += f"\n{len(self.subspecies)} Subspecies: "
            for subspecies in self.subspecies:
                return_str += f"{subspecies.name}, "
        return_str += f"\nOrder: {self.order}"
        return_str += f"\nFamily: {self.family}: {self['Family_English_name']}"
        try:
            return_str += f"\nRange: {self['Range']}"
        except KeyError:
            return_str += "\nRange Not Specified"
        return return_str + '\n'



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 22 11:29:58 2025

@author: Thomas Lee
Rice University
Department of Earth, Environmental, and Planetary Sciences
Email: tl165@rice.edu
"""
all = ['database']
DATA_FILES = ["AviList-v2025-11Jun-extended.xlsx","AviListDataBase.pickle"]


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 22 14:25:46 2025

@author: Thomas Lee
Rice University
Department of Earth, Environmental, and Planetary Sciences
Email: tl165@rice.edu
"""

from AviListPy.taxonomy.family import Family, Genus, Species
from AviListPy.data.avilistdatabase import AviListDataBase
from typing import Any, KeysView, ValuesView, ItemsView, Iterator, List
from pandas import DataFrame

class Order():
    """Container for a Family in the AviList DataBase

    The highest taxonomic rank in the AviList.taxonomy class system. Container
    for all family objects corresponding with family
    in this order. See Order.keys() for a list of all available data. Can
    access data in a dictionary-like manner where keys are columns in the
    AviList, but iterating will iterate through Order.families

    Attributes:
    -----------
    db: AviList.data.avilistdatabase.AviListDataBase
        AviListDataBase class. It is recommended to pass an existing
        AviListDataBase object to the Family class during initialization,
        but if none is given it will initialize one from the Excel sheet. See
        the Setup section on the main GitHub page for more detail.
    df: Pandas.DataFrame
        The single row for this family in AviList as a Pandas DataFrame.
    name: str
        Scientific name for this family, from self['Scientific_name']
    order: str
        Taxonomic order
    family: str
        Taxonomic family
    genera: list of AviList.taxonomy.genus.Genus
        All genera within this family.
    species: list of AviList.taxonomy.species.Species
        All species within this family.

    Example:
        >>> db = AviListDataBase()
        >>> order = Order('Anseriformes',db=db)
        >>> order.name
        'Anseriformes'
        >>> order['Sequence']
        230

    """
    def __init__(self, name: str, exact: bool=False, load_subspecies: bool=False, db: AviListDataBase=None):
        """
        Parameters
        ----------
        name : str
            Scientific name of the species to search for.
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any English name
            in the database, and is not case sensitive. The default is False.
        load_subspecies: bool, optional
            If True, will load Subspecies objects while loading Species. See
            AviList.taxonomy.species.Species for more information.
        db : AviListDataBase, optional
            AviListDataBase. The default is None.
        """
        if db is None:
            self.db = AviListDataBase()
        else:
            self.db = db
        self.df = self.lookup_order(name)
        self._data = self.df.iloc[0].to_dict()
        self.name = self._data['Scientific_name']
        self.families = self.find_matching_families(load_subspecies=load_subspecies)
        self.genera = self.find_matching_genera()
        self.species = self.find_matching_species()

    def __str__(self) -> str:
        return_str = f'{self.name}'
        num_equals = (80 - len(return_str)) // 2
        return_str = '='*num_equals + return_str + '='*num_equals
        return_str += f'\n{len(self.species)} species in {len(self.genera)} genera across {len(self.families)} families'
        for key, val in self.items():
            return_str += (f'\n{key}: {val}')
        return return_str + '\n'

    def __iter__(self) -> Iterator[Any]:
        return iter(self.families)

    def __len__(self) -> int:
        return len(self.families)

    def __getitem__(self, key) -> Any:
        return self._data[key]

    def __setitem__(self, key, value) -> None:
        self._data[key] = value

    def __contains__(self, key) -> bool:
        return key in self._data

    def keys(self) -> KeysView:
        """Returns keys in a dictionary.keys() like manner"""
        return self._data.keys()

    def values(self) -> ValuesView:
        """Returns values in a dictionary.values() like manner"""
        return self._data.values()

    def items(self) -> ItemsView:
        """Returns keys, values in a dictionary.items() like manner"""
        return self._data.items()

    def lookup_order(self, name) -> DataFrame:
        """
        Parameters
        ----------
        name : str
            Order to search for.
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any scientific name
            in the database, and is not case sensitive. The default is False.

        Returns
        -------
        _order_df : pandas.DataFrame
           Pandas DataFrame with only one row containing the entry for the order.

        """
        df = self.db.df[self.db.df['Taxon_rank'].str.contains('order')]
        _family_df = df[df['Scientific_name'].str.contains(name, case=False, na=False)]
        if _family_df.shape[0] == 0:
            raise ValueError('No matching families found')
        if _family_df.shape[0] > 1:
            fail_str = f'{name} could refer to: \n'
            for _family_ in _family_df['Scientific_name'].to_list():
                fail_str += (f'{_family_}, ')
            raise ValueError(fail_str)
        _family_df = _family_df.dropna(axis=1)
        return _family_df

    def find_matching_families(self, load_subspecies: bool=False) -> List[Family]:
        """
        Parameters
        ----------
        exact : bool, optional
            If True, will only search for an exact match for the name string.
            If False, searches for name as a substring of any scientific name
            in the database, and is not case sensitive. The default is False.
        load_subspecies : bool, optional
            If True, loads subspecies. The default is False.

        Returns
        -------
        matching_family_list : list of AviList.taxonomy.family.Family
            List of families contained within the order.
        """
        family_df = self.db.df[self.db.df['Taxon_rank'] == 'family']
        matching_family_df = family_df[family_df['Order'] == str(self.name)]
        print(f'Loading {len(matching_family_df)} families in order {self.name}')
        if len(matching_family_df) == 0:
            raise ValueError('No matching families found')
        matching_family_list = []
        for _matching_family_name in matching_family_df['Scientific_name'].to_list():
            matching_family_list.append(Family(_matching_family_name, db=self.db, exact=True, load_subspecies=load_subspecies))
        return matching_family_list

    def find_matching_genera(self) -> List[Genus]:
        """Returns a list of all genera in this order as Genus objects"""
        genera_list = []
        for family in self.families:
            genera_list += family.genera
        return genera_list

    def find_matching_species(self) -> List[Species]:
        """Returns a list of all species in this order as Species objects"""
        species_list = []
        for genus in self.genera:
            species_list += genus.species
        return species_list

    def show_families(self) -> None:
        print(f'{len(self.families)} family in order {self.name}')
        count = 0
        for family in self.families:
            print(f'{family.name}: {len(family.genera)} genera, {len(family.species)} species')
            count += len(family.species)
        print(f'{count} total species in {self.name}')

    def show_genera(self) -> None:
        print(f'{len(self.genera)} genera in family {self.name}')
        count = 0
        for genus in self.genera:
            print(f'{genus.name}: {len(genus.species)} species')
            count += len(genus.species)
        print(f'{count} total species in {self.name}')


# -*- coding: utf-8 -*-
"""
AviListPy
==============================================================================
Python implementation of the 2025 AviList Global Avian Checklist.

@author: Thomas Lee
Rice University
Department of Earth, Environmental, and Planetary Sciences
Email: tl165@rice.edu

Citations:
AviList Core Team. 2025. AviList: The Global Avian Checklist, v2025.
https://doi.org/10.2173/avilist.v2025
"""
from pathlib import Path

# Package metadata
__author__ = 'Thomas Lee'
__email__ = 'tl165@rice.edu'
__license__ = 'CC0-1.0'

def _get_version():
    """Read version from pyproject.toml file."""
    try:
        import tomllib  # Python 3.11+
    except ImportError:
        try:
            import tomli as tomllib  # Fallback for older Python
        except ImportError:
            # Final fallback - try to parse manually or return default
            return "unknown"

    try:
        pyproject_path = Path(__file__).parent.parent / "pyproject.toml"
        with open(pyproject_path, "rb") as f:
            pyproject_data = tomllib.load(f)
        return pyproject_data["project"]["version"]
    except (FileNotFoundError, KeyError, Exception):
        # Fallback version if reading fails
        return "unknown"

__version__ = _get_version()

from AviListPy.data.avilistdatabase import AviListDataBase
from AviListPy.taxonomy.species import Species
from AviListPy.taxonomy.genus import Genus
from AviListPy.taxonomy.family import Family
from AviListPy.taxonomy.order import Order
from AviListPy.taxonomy.subspecies import Subspecies

__all__ = ['AviListDataBase', 'Order', 'Family', 'Genus', 'Species', 'Subspecies']
